package net.puppygames.tools;

import java.nio.DoubleBuffer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Random;
import java.util.concurrent.CopyOnWriteArrayList;

import org.lwjgl.BufferUtils;
import org.lwjgl.input.Keyboard;

import riven.classpath.math.Matrix4;
import riven.classpath.math.Plane;
import riven.classpath.math.Vec3;

import static org.lwjgl.opengl.GL11.*;

public class RoomsDemo extends DemoFramework3D {
	public static void main(String[] args) {
		new RoomsDemo().mainLoop();
	}

	@Override
	protected void input() {
		super.input();

		{
			// bounce off backsides of portals
			playerPosNew.add(movement);
			for (Portal portal : currentUniverse.portals) {
				if (playerPosNew.distance(portal.origin) < portal.width * 0.5f) {
					if (portal.side(playerPosOld) < 0.0f) {
						if (portal.side(playerPosNew) > 0.0f) {
							// back to previous position
							playerPosNew.load(playerPosOld);
						}
					}
				}
			}
			playerPosOld.load(playerPosNew);

			cameraPosition.load(playerPosNew).y += 1.75f;
		}

		// move from one universe to the other, if the portal is moved through in the right direction
		{
			Vec3 cam = new Vec3(cameraPosition).mul(1, 0, 1);
			Portal nearestPortal = null;
			for (Portal portal : currentUniverse.portals) {
				if (cam.distance(portal.origin) < portal.width * 0.5f) {
					nearestPortal = portal;
					break;
				}
			}

			if (nearestPortal == null) {
				portalState = STATE_AWAY;
			} else {
				if (portalState == STATE_AWAY) {
					if (nearestPortal.side(cam) > 0.0f) {
						portalState = STATE_INFRONT;
						System.out.println("in front of " + nearestPortal.toUniverse);
					}
				} else if (portalState == STATE_INFRONT) {
					if (nearestPortal.side(cam) < 0.0f) {
						portalState = STATE_BEHIND;
						currentUniverse = nearestPortal.toUniverse;
						portalState = STATE_INFRONT;
						System.out.println("entered " + nearestPortal.toUniverse);
					}
				}
			}
		}
	}

	@Override
	protected void onKeyboardEvent() {
		if (Keyboard.getEventKey() == Keyboard.KEY_SPACE) {
			if (Keyboard.getEventKeyState()) {
				System.out.println("new portal");

				for (Portal portal : currentUniverse.portals) {
					portal.remove();
				}

				Universe to;
				do {
					to = universes[(int) (Math.random() * universes.length)];
				} while (to == currentUniverse);

				Matrix4 mat = new Matrix4();
				mat.translate(playerPosNew);
				mat.rotY(cameraAngleH);
				mat.translate(new Vec3(0, 0, +5)); // in front of player
				Vec3 origin = mat.transform(new Vec3());

				currentUniverse.addTwoWayPortal(origin, cameraAngleH, to);
			}
		}
	}

	private class Universe {
		public int index, wallHeight;
		public Vec3 wallColor, skyColor;
		public List<Portal> portals = new CopyOnWriteArrayList<>();
		public List<Runnable> renderOps = new ArrayList<>();

		public Universe(int index, Vec3 wallColor, Vec3 skyColor) {
			this.index = index;
			this.wallColor = wallColor;
			this.skyColor = skyColor;
		}

		public void addTwoWayPortal(Vec3 origin, float angle, Universe to) {
			Portal p1 = new Portal(origin, 1.5f, 2.5f, this, to, angle);
			Portal p2 = new Portal(origin, 1.5f, 2.5f, to, this, angle + 180.0f);
			p1.copy = p2;
			p2.copy = p1;
		}

		@Override
		public String toString() {
			return "Universe " + (char) ('A' + index);
		}
	}

	private class Portal {
		Vec3 origin;
		float width, height;
		Universe fromUniverse, toUniverse;
		float angle;
		Portal copy;

		public Portal(Vec3 origin, float width, float height, Universe fromUniverse, Universe toUniverse, float angle) {
			this.origin = origin;
			this.width = width;
			this.height = height;
			this.fromUniverse = fromUniverse;
			this.toUniverse = toUniverse;
			this.fromUniverse.portals.add(this);
			this.angle = angle;
		}

		public float side(Vec3 cam) {
			Matrix4 mat = new Matrix4();
			mat.translate(origin);
			mat.rotY(angle);
			Vec3 a = mat.transform(new Vec3(+1, 0, 0));
			Vec3 b = mat.transform(new Vec3(-1, 0, 0));
			return RoomsDemo.this.side(a, b, cam);
		}

		public Plane plane() {
			Vec3 a = new Vec3(+1, 0, 0);
			Vec3 b = new Vec3(-1, 0, 0);
			Vec3 c = new Vec3(0, +1, 0);

			Matrix4 mat = new Matrix4();
			mat.translate(origin);
			mat.rotY(angle);
			mat.transform(a);
			mat.transform(b);
			mat.transform(c);
			return Plane.fromTri(a, b, c);
		}

		public void remove() {
			if (this.fromUniverse.portals.remove(this)) {
				if (copy != null) {
					copy.remove();
				}
			}
		}
	}

	private Universe currentUniverse = null;
	private final Universe[] universes = new Universe[4];
	{
		Random rndm2 = new Random(123);
		Random rndm3 = new Random(123456 + 1);

		for (int i = 0; i < universes.length; i++) {
			Vec3 wallColor = new Vec3(rndm3.nextFloat(), rndm3.nextFloat(), rndm3.nextFloat());
			Vec3 skyColor;
			if (i == 0 || i == 2) {
				skyColor = new Vec3(0.75f, 0.85f, 1.00f);
			} else {
				skyColor = new Vec3(0.33f, 0.05f, 0.00f);
			}
			universes[i] = new Universe(i, wallColor, skyColor);
			universes[i].wallHeight = 1 + rndm2.nextInt(5) + (i == 0 ? 2 : 0);

			if (i == 2) {
				universes[i].renderOps.add(SNOW_RENDER_OP);
			} else if (i == 3) {
				universes[i].renderOps.add(LAVA_RAIN_RENDER_OP);
			}
		}

		currentUniverse = universes[0];
	}

	private static final int STATE_AWAY = 0;
	private static final int STATE_INFRONT = 1;
	private static final int STATE_BEHIND = 2;
	private int portalState = STATE_AWAY;

	@Override
	protected void render() {
		{
			drawUniverse(16, currentUniverse.wallHeight, currentUniverse, true);
		}

		// depth sort portals! (back-to-front)
		List<Portal> sorted = new ArrayList<Portal>(currentUniverse.portals);
		Collections.sort(sorted, new Comparator<Portal>() {
			@Override
			public int compare(Portal o1, Portal o2) {
				float d1 = new Vec3(cameraPosition).mul(1, 0, 1).distance(o1.origin);
				float d2 = new Vec3(cameraPosition).mul(1, 0, 1).distance(o2.origin);
				return d1 > d2 ? -1 : +1;
			}
		});

		glColorMask(false, false, false, false);
		glEnable(GL_STENCIL_TEST);

		// draw stencils
		for (Portal portal : sorted) {
			// can we even look into the portal?
			if (portal.side(cameraPosition) < 0.0) {
				//continue;
			}

			int bit = 1 << (portal.toUniverse.index);

			// always accept fragments (with target=stencil) 
			glStencilFunc(GL_ALWAYS, bit, bit);

			// replace whatever is in the stencil buffer, except on z-fail
			glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);

			glDisable(GL_CULL_FACE);
			glDepthMask(false);

			// frontside portal as stencil
			glPushMatrix();
			glTranslatef(portal.origin.x, portal.origin.y, portal.origin.z);
			glRotatef(portal.angle, 0, 1, 0);
			drawPortalSurface(portal.width, portal.height);
			glPopMatrix();

			glDepthMask(true);
			glEnable(GL_CULL_FACE);
		}

		// ---

		final float doorHolder = 0.025f;

		//	glStencilMask(0x00);// which bits to write into

		for (Portal portal : sorted) {
			// can we even look into the portal?
			if (portal.side(cameraPosition) < 0.0) {
				//continue;
			}

			int bit = 1 << (portal.toUniverse.index);

			// draw only where stencil's value is set
			glStencilFunc(GL_EQUAL, bit, bit); // which bits to read from	
			glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

			glFrontFace(GL_CCW);
			{
				// ensure depth-buffer at stencil == portal surface
				glColorMask(false, false, false, false);
				glPushMatrix();
				glTranslatef(portal.origin.x, portal.origin.y, portal.origin.z);
				glRotatef(portal.angle, 0, 1, 0);
				drawPortalSurface(portal.width, portal.height);
				glPopMatrix();
				glColorMask(true, true, true, true);
			}

			boolean doClipPortal = true;
			if (doClipPortal) {
				Plane plane = portal.plane();
				//modelViewMatrix.transform(plane);

				DoubleBuffer planeData = BufferUtils.createDoubleBuffer(4);
				planeData.put(plane.x);
				planeData.put(plane.y);
				planeData.put(plane.z);
				planeData.put(plane.w);
				planeData.flip();

				glClipPlane(GL_CLIP_PLANE0, planeData);

				glEnable(GL_CLIP_PLANE0);
			}
			{
				// render new universe, in stencil
				drawUniverse(16, portal.toUniverse.wallHeight, portal.toUniverse, false);
			}
			if (doClipPortal) {
				glDisable(GL_CLIP_PLANE0);
			}

			// render portal border on top of portal
			glFrontFace(GL_CCW);
			glColor(new Vec3(portal.toUniverse.wallColor).mul(0.5f));
			glPushMatrix();
			glTranslatef(portal.origin.x, portal.origin.y, portal.origin.z);
			glRotatef(portal.angle, 0, 1, 0);
			drawPortalStructure(portal.width, portal.height, doorHolder);
			glPopMatrix();
		}

		glDisable(GL_STENCIL_TEST);
	}

	private float side(Vec3 a, Vec3 b, Vec3 p) {
		return (b.x - a.x) * (p.z - a.z) - (b.z - a.z) * (p.x - a.x);
	}

	private void drawPortalSurface(float w, float h) {
		glBegin(GL_QUADS);
		drawDoorAlignedQuad(-w * 0.5f, 0, w, h);
		glEnd();
	}

	private void drawPortalStructure(float w, float h, float margin) {
		glBegin(GL_QUADS);
		drawDoorAlignedQuad(-w * 0.5f + margin * 0, 0, margin, h);
		drawDoorAlignedQuad(+w * 0.5f - margin * 1, 0, margin, h);
		glEnd();

		glPushMatrix();
		{
			glTranslatef(0, h - margin, 0);
			glBegin(GL_QUADS);
			drawDoorAlignedQuad(-w * 0.5f + margin, 0, w - margin * 2, margin);
			glEnd();
		}
		glPopMatrix();
	}

	private void drawDoorAlignedQuad(float x, float z, float w, float h) {
		glVertex3f(x + w * 0, 0, z);
		glVertex3f(x + w * 1, 0, z);
		glVertex3f(x + w * 1, h, z);
		glVertex3f(x + w * 0, h, z);
	}

	private void drawUniverse(int dim, int wallHeight, Universe universe, boolean tripleFlipDeluxe) {

		Vec3 color = universe.wallColor;

		// skybox
		{
			glDepthFunc(GL_ALWAYS);

			glFrontFace(GL_CCW);
			glColor(universe.skyColor);
			renderSky(50.0f);

			glDepthFunc(GL_LEQUAL);
		}

		// floor
		{
			glFrontFace(GL_CCW);
			drawTiles(1.0f, dim, dim, color);
		}

		// walls
		{
			{
				glFrontFace(GL_CW);
				glPushMatrix();
				glRotatef(-90.0f, 1, 0, 0);
				drawTiles(1.0f, dim, wallHeight, color);
				glPopMatrix();
			}

			{
				glFrontFace(GL_CCW);
				glPushMatrix();
				glTranslatef(0, 0, dim);
				glRotatef(-90.0f, 1, 0, 0);
				drawTiles(1.0f, dim, wallHeight, color);
				glPopMatrix();
			}

			{
				glFrontFace(GL_CW);
				glPushMatrix();
				glRotatef(-90.0f, 0, 0, 1);
				glScalef(-1, 1, 1);
				drawTiles(1.0f, wallHeight, dim, color);
				glPopMatrix();
			}

			{
				glFrontFace(GL_CCW);
				glPushMatrix();
				glTranslatef(dim, 0, 0);
				glRotatef(-90.0f, 0, 0, 1);
				glScalef(-1, 1, 1);
				drawTiles(1.0f, wallHeight, dim, color);
				glPopMatrix();
			}
		}

		glFrontFace(tripleFlipDeluxe ? GL_CW : GL_CCW);

		if (universe.index == 1) {
			Random rndm5 = new Random(36 + 1 + universe.index);
			glBegin(GL_QUADS);
			for (int x = 0; x < 16; x++) {
				for (int z = 0; z < 16; z++) {
					glColor3f(rndm5.nextFloat(), rndm5.nextFloat(), rndm5.nextFloat());
					renderCube(x + 0.1f, 0 + 0.1f, z + 0.1f, 0.8f, rndm5.nextFloat() * 0.8f, 0.8f, -1);
				}
			}
			glEnd();
		}

		// backside of portals
		for (Portal portal : universe.portals) {
			if (portal.side(cameraPosition) > 0.0) {
				//continue;
			}

			if (portal.toUniverse == universe) {
				//continue;
			}

			glColor(new Vec3(portal.toUniverse.wallColor).mul(0.5f));

			glPushMatrix();
			glTranslatef(portal.origin.x, portal.origin.y, portal.origin.z);
			glRotatef(portal.angle, 0, 1, 0);
			drawPortalSurface(portal.width, portal.height);
			glPopMatrix();
		}

		for (Portal portal : universe.portals) {
			if (portal.toUniverse == universe) {
				//continue;
			}

			glColor(new Vec3(portal.toUniverse.wallColor).mul(0.5f));

			glPushMatrix();
			glTranslatef(portal.origin.x, portal.origin.y, portal.origin.z);
			glRotatef(portal.angle, 0, 1, 0);
			glTranslatef(-0.5f * portal.width, 0.0f, 0.0f);
			renderCube(portal.width, portal.height, 0.025f);
			glPopMatrix();
		}

		glFrontFace(GL_CW);

		for (Runnable renderOp : universe.renderOps) {
			renderOp.run();
		}
	}

	private void renderSky(float s) {
		Vec3 nw1 = new Vec3(+s, -s, -s);
		Vec3 ne1 = new Vec3(+s, -s, +s);
		Vec3 sw1 = new Vec3(-s, -s, -s);
		Vec3 se1 = new Vec3(-s, -s, +s);
		Vec3 nw2 = new Vec3(+s, +s, -s);
		Vec3 ne2 = new Vec3(+s, +s, +s);
		Vec3 sw2 = new Vec3(-s, +s, -s);
		Vec3 se2 = new Vec3(-s, +s, +s);

		glBegin(GL_QUADS);
		{
			// bottom
			glVertex(nw1);
			glVertex(ne1);
			glVertex(se1);
			glVertex(sw1);

			// north
			glVertex(nw2);
			glVertex(ne2);
			glVertex(ne1);
			glVertex(nw1);

			// south
			glVertex(sw1);
			glVertex(se1);
			glVertex(se2);
			glVertex(sw2);

			// east
			glVertex(ne2);
			glVertex(se2);
			glVertex(se1);
			glVertex(ne1);

			// west
			glVertex(nw1);
			glVertex(sw1);
			glVertex(sw2);
			glVertex(nw2);

			// top
			glVertex(sw2);
			glVertex(se2);
			glVertex(ne2);
			glVertex(nw2);
		}
		glEnd();
	}

	private void renderCube(float w, float h, float d) {
		Vec3 nw1 = new Vec3(w, 0, 0);
		Vec3 ne1 = new Vec3(w, 0, d);
		Vec3 sw1 = new Vec3(0, 0, 0);
		Vec3 se1 = new Vec3(0, 0, d);
		Vec3 nw2 = new Vec3(w, h, 0);
		Vec3 ne2 = new Vec3(w, h, d);
		Vec3 sw2 = new Vec3(0, h, 0);
		Vec3 se2 = new Vec3(0, h, d);

		glBegin(GL_QUADS);
		{
			// bottom
			glVertex(nw1);
			glVertex(ne1);
			glVertex(se1);
			glVertex(sw1);

			// north
			glVertex(nw2);
			glVertex(ne2);
			glVertex(ne1);
			glVertex(nw1);

			// south
			glVertex(sw1);
			glVertex(se1);
			glVertex(se2);
			glVertex(sw2);

			// east
			glVertex(ne2);
			glVertex(se2);
			glVertex(se1);
			glVertex(ne1);

			// west
			if (false) {
				glVertex(nw1);
				glVertex(sw1);
				glVertex(sw2);
				glVertex(nw2);
			}

			// top
			glVertex(sw2);
			glVertex(se2);
			glVertex(ne2);
			glVertex(nw2);
		}
		glEnd();
	}

	private void drawTiles(float floorDim, int xRepeat, int zRepeat, Vec3 color) {
		Random r = new Random(xRepeat * zRepeat);
		glBegin(GL_QUADS);
		for (int z = 0; z < zRepeat; z++) {
			for (int x = 0; x < xRepeat; x++) {

				float shade = r.nextFloat() * 0.05f;
				glColor3f(color.x - shade, color.y - shade, color.z - shade);

				glVertex3f(floorDim * (x + 0), 0.0f, floorDim * (z + 0));
				glVertex3f(floorDim * (x + 1), 0.0f, floorDim * (z + 0));
				glVertex3f(floorDim * (x + 1), 0.0f, floorDim * (z + 1));
				glVertex3f(floorDim * (x + 0), 0.0f, floorDim * (z + 1));
			}
		}
		glEnd();
	}

	private static final Runnable SNOW_RENDER_OP = new Runnable() {
		Random r = new Random(2345);
		Vec3[] pos = new Vec3[8 * 1024];
		Vec3[] vel = new Vec3[pos.length];
		{
			for (int i = 0; i < pos.length; i++) {
				pos[i] = new Vec3();
				vel[i] = new Vec3();

				init(i);
			}
		}

		private void init(int i) {
			pos[i].x = r.nextFloat() * 16;
			pos[i].y = r.nextFloat() * 10;
			pos[i].z = r.nextFloat() * 16;

			vel[i].x = (r.nextFloat() - 0.5f) * 2.0f;
			vel[i].y = 0.75f + r.nextFloat() * 0.25f;
			vel[i].z = (r.nextFloat() - 0.5f) * 2.0f;

			vel[i].mul(ONE_UNIT_PER_SECOND);
			vel[i].mul(0.5f, -0.5f, 0.5f);
		}

		@Override
		public void run() {
			glPointSize(2.0f);

			glColor3f(1, 1, 1);
			glBegin(GL_POINTS);

			for (int i = 0; i < pos.length; i++) {
				pos[i].add(vel[i]);

				if (pos[i].y < 0.0f) {
					this.init(i);
				} else if (pos[i].y < 0.01f) {
					vel[i].load(0, -0.005f, 0).mul(ONE_UNIT_PER_SECOND); // thaw
				}

				glVertex(pos[i]);
			}

			glEnd();
		}
	};

	private static final Runnable LAVA_RAIN_RENDER_OP = new Runnable() {
		Random r = new Random(2345);
		Vec3[] pos = new Vec3[8 * 1024];
		Vec3[] vel = new Vec3[pos.length];
		Vec3[] acc = new Vec3[pos.length];
		{
			for (int i = 0; i < pos.length; i++) {
				pos[i] = new Vec3();
				vel[i] = new Vec3();
				acc[i] = new Vec3();

				init(i);
			}
		}

		private void init(int i) {
			pos[i].x = r.nextFloat() * 16;
			pos[i].y = r.nextFloat() * 10;
			pos[i].z = r.nextFloat() * 16;

			vel[i].x = 0.75f + r.nextFloat() * 0.25f;
			vel[i].y = 0.90f + r.nextFloat() * 0.10f;
			vel[i].z = 0.75f + r.nextFloat() * 0.25f;
			vel[i].mul(ONE_UNIT_PER_SECOND);
			vel[i].mul(0.5f, -5.5f, 0.1f);

			acc[i].y -= 0.1f;
			acc[i].mul(ONE_UNIT_PER_SECOND);
		}

		@Override
		public void run() {
			glPointSize(2.0f);

			glColor3f(1.25f, 1.25f, 0.50f);
			glBegin(GL_POINTS);

			for (int i = 0; i < pos.length; i++) {
				vel[i].add(acc[i]);
				pos[i].add(vel[i]);

				if (pos[i].x > 16 || pos[i].z > 16) {
					this.init(i);
				} else if (pos[i].y <= 0.0f) {
					if (vel[i].y < -0.1f) {
						pos[i].y = 0.001f;
						vel[i].mul(0.15f, -0.05f, 0.15f);
					} else {
						this.init(i);
					}
				}

				glVertex(pos[i]);
			}

			glEnd();
		}
	};
}
